/*
 * The contents of this file are subject to the terms
 * of the Common Development and Distribution License
 * (the License).  You may not use this file except in
 * compliance with the License.
 *
 * You can obtain a copy of the license at
 * https://glassfish.dev.java.net/public/CDDLv1.0.html or
 * glassfish/bootstrap/legal/CDDLv1.0.txt.
 * See the License for the specific language governing
 * permissions and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL
 * Header Notice in each file and include the License file
 * at glassfish/bootstrap/legal/CDDLv1.0.txt.
 * If applicable, add the following below the CDDL Header,
 * with the fields enclosed by brackets [] replaced by
 * you own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 */

package com.sun.enterprise.web.portunif.util;

import com.sun.enterprise.web.connector.grizzly.DefaultReadTask;
import com.sun.enterprise.web.connector.grizzly.OutputWriter;
import com.sun.enterprise.web.connector.grizzly.Pipeline;
import com.sun.enterprise.web.connector.grizzly.ReadTask;
import com.sun.enterprise.web.connector.grizzly.SecureSelector;
import com.sun.enterprise.web.connector.grizzly.SelectorThread;
import com.sun.enterprise.web.connector.grizzly.StreamAlgorithm;
import com.sun.enterprise.web.connector.grizzly.algorithms.NoParsingAlgorithm;
import com.sun.enterprise.web.connector.grizzly.ssl.SSLOutputWriter;
import com.sun.enterprise.web.connector.grizzly.ssl.SSLPipeline;
import com.sun.enterprise.web.connector.grizzly.ssl.SSLReadTask;
import com.sun.enterprise.web.connector.grizzly.ssl.SSLSelectorThread;
import java.io.IOException;
import java.net.InetAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.net.ssl.SSLEngineResult;
import javax.net.ssl.SSLException;

/**
 * Utility class used to redirect an HTTP(s) request to another protocol and
 * endpoint. The following scenarios are supported:
 *
 * (1) http://host:port1 to https://host:port1
 * (2) https://host:port1 to http://host:port1
 * (3) http://host:port2 to https://host:port1
 * (4) https://host:port2 to https://host:port1
 * (5) http://host:port2 to http://host:port1
 * (6) https://host:port2 to http://host:port1
 *
 * This class internally start an NIO Selector listening on an 'external' port
 * to a 'redirect' port. All requests to the external port,
 * independently of the protocol are redirected to the 'redirect' port.
 *
 * @author Jeanfrancois Arcand
 */
public class Redirector {
    
    private final static String WRONG_PROTOCOL =
            "Unrecognized SSL message, plaintext connection?";
    
    /**
     * The port on which all request will be redirected.
     */
    private int redirectedPort = 4848;
    
    
    /**
     * The 'external' port where all request will be redirected to the
     * 'redirectedPort'
     */
    private int externalPort = 4849;
    
    
    /**
     * <tt>true</tt> if the 'redirected' port listen to https, <tt>false</tt>
     * is it listen to http.
     */
    private boolean isRequestedTransportSecure = false;
    
    
    /**
     * HTTP end line.
     */
    private final static String NEWLINE = "\r\n";
    
    /**
     * The IP address the server is running on.
     */
    private static String ipAddress = "127.0.0.1";

    
    /**
     * Is the buffer initialized.
     */
    private boolean initialized = false;
    
    
    static {
        try{
            ipAddress = InetAddress.getLocalHost().getHostAddress();
        } catch (Throwable t){
            ;
        }
    }
    
    
    /**
     * Http to Https ByteBuffer response.
     */
    private ByteBuffer httpToHttps;
    
    
    /**
     * Https to Https ByteBuffer response.
     */
    private ByteBuffer httpsToHttps;
    
    /**
     * Https to Http ByteBuffer response.
     */
    private ByteBuffer httpsToHttp;
    
    
    /**
     * Http to Http ByteBuffer response.
     */
    private ByteBuffer httpToHttp;
    
    
    /**
     * HTTP SC_FOUND header.
     */
    private final static ByteBuffer SC_FOUND =
            ByteBuffer.wrap( ("HTTP/1.1 302" + NEWLINE).getBytes());
    
    
    /**
     * The <code>Selector</code> timeout value.
     */
    protected int selectorTimeout = 60000;
    
    
    // --------------------------------------------------------- Constructor --/
    
    
    public Redirector(){
    }
    
    
    public Redirector(int redirectedPort, int externalPort){
        this.redirectedPort = redirectedPort;
        this.externalPort = externalPort;
    }
    
    
    private void initializeBuffer(){
        httpToHttps =
                ByteBuffer.wrap(
                (new String("Location: https://" + ipAddress + ":" + redirectedPort
                + NEWLINE + "Connection:close" + NEWLINE
                + "Cache-control: private" + NEWLINE
                + NEWLINE).getBytes()));
        
        httpsToHttps =
                ByteBuffer.wrap(
                (new String("Location: https://" + ipAddress + ":" + redirectedPort
                + NEWLINE + "Connection:close" + NEWLINE
                + "Cache-control: private" + NEWLINE
                + NEWLINE).getBytes()));
        
        httpsToHttp =
                ByteBuffer.wrap(
                (new String("Location: http://" + ipAddress + ":" + redirectedPort
                + NEWLINE + "Connection:close" + NEWLINE
                + "Cache-control: private" + NEWLINE
                + NEWLINE).getBytes()));
        
        httpToHttp =
                ByteBuffer.wrap(
                (new String("Location: http://" + ipAddress + ":" +  redirectedPort
                + NEWLINE + "Connection:close" + NEWLINE
                + "Cache-control: private" + NEWLINE
                + NEWLINE).getBytes()));       
        
        initialized = true;
    }
    
    
    
    /**
     * Start the 'externalPort' Selector. All request to this Selector will
     * be redirected to the 'redirectedPort'.
     */
    public void start(){
        try{
                            
            if (!initialized){
                initializeBuffer();
            }
            final SSLSelectorThread sslSelector = new SSLSelectorThread(){
                
                private ProtocolInfo protocolInfo = new ProtocolInfo();
                
                /**
                 * Create a new <code>Pipeline</code> instance using the
                 * <code>pipelineClassName</code> value.
                 */
                protected Pipeline newPipeline(int maxThreads,
                        int minThreads,
                        String name,
                        int port,
                        int priority){
                    
                    Pipeline pipeline = new SSLPipeline();
                    pipeline.setMaxThreads(maxThreads);
                    pipeline.setMinThreads(minThreads);
                    pipeline.setName(name);
                    pipeline.setPort(port);
                    pipeline.setPriority(priority);
                    pipeline.setQueueSizeInBytes(getMaxQueueSizeInBytes());
                    pipeline.setThreadsIncrement(getThreadsIncrement());
                    pipeline.setThreadsTimeout(getThreadsTimeout());
                    
                    return pipeline;
                    
                }
                
                
                protected DefaultReadTask newReadTask(){
                    StreamAlgorithm streamAlgorithm = new NoParsingAlgorithm();
                    streamAlgorithm.setPort(getPort());
                    
                    SSLReadTask task = new SSLReadTask(){
                        public void doTask() throws IOException {
                            int count = 0;
                            Exception exception = null;
                            boolean keepAlive = false;
                            SSLEngineResult result;
                            
                            SocketChannel socketChannel =
                                    (SocketChannel)key.channel();
                            try {
                                allocateBuffers();
                                keepAlive = false;
                                try{
                                    if ( doHandshake() ) {
                                        protocolInfo.protocol = "https";
                                        protocolInfo.sslEngine = sslEngine;
                                        protocolInfo.outputBB = outputBB;
                                        protocolInfo.isSecure = true;
                                    }
                                } catch(SSLException ex){
                                    if (ex.getMessage().equals(WRONG_PROTOCOL)){
                                        protocolInfo.protocol = "http";
                                        protocolInfo.isSecure = false;
                                    }
                                } finally {
                                    protocolInfo.socketChannel = socketChannel;
                                    protocolInfo.isRequestedTransportSecure =
                                            isRequestedTransportSecure;
                                    protocolInfo.byteBuffer = byteBuffer;
                                    if ( protocolInfo.isSecure ) {
                                        redirectSSL(protocolInfo);
                                    } else {
                                        redirect(protocolInfo);
                                    }
                                }
                            } catch (IOException ex) {
                                log(ex);
                                exception = ex;
                            } catch (RuntimeException ex) {
                                exception = ex;
                            } finally {
                                manageKeepAlive(keepAlive,count,exception);
                            }
                        }
                        
                    };
                    task.setPipeline(readPipeline);
                    task.setSelectorThread(this);
                    task.setRecycle(isRecycleTasks());
                    task.initialize(streamAlgorithm,
                            isUseDirectByteBuffer(), isUseByteBufferView());
                    
                    return task;
                }
                
                protected ReadTask handleRead(SelectionKey key)
                    throws IOException{
                    
                    if ( sslContext == null ){
                        Enumeration<SelectorThread> selectors
                                = SelectorThread.getSelectors();
                        SelectorThread sel;
                        while (selectors.hasMoreElements()){
                            sel = selectors.nextElement();
                            if (sel instanceof SSLSelectorThread){
                                sslContext =
                                        ((SSLSelectorThread)sel).getSSLContext();
                                if (sslContext != null)
                                    break;
                            }
                        }
                        
                        if (sslContext == null)
                            throw new
                               RuntimeException("SSL protocol not supported");
                        
                        SelectorThread targettedSelector =
                                SelectorThread.getSelector(redirectedPort);
                        // No other way than instanceof
                        if ( targettedSelector instanceof SecureSelector ){
                            isRequestedTransportSecure = true;
                        }
                    }
                    return super.handleRead(key);
                }
            };
            sslSelector.setPort(externalPort);
            
            new Thread(){
                public void run(){
                    try{
                        sslSelector.initEndpoint();
                        sslSelector.startEndpoint();
                    } catch (Throwable ex){
                        log(ex);
                    }
                }
            }.start();
        } catch (Throwable t){
            log(t);
        }
    }
    
    
    /**
     * Redirect a secure request (https) to http or https.
     * @param protocolInfo the ProtocolInfo that contains the information about
     *               the current protocol state.
     */
    public final void redirectSSL(ProtocolInfo protocolInfo) throws IOException{
                        
        if (!initialized){
            initializeBuffer();
        }
                
        redirectSSL(protocolInfo,
                protocolInfo.isRequestedTransportSecure ? httpsToHttps: httpsToHttp);
    }
    
    
    /**
     * Redirect a secure request (https) to http or https.
     * @param protocolInfo the ProtocolInfo that contains the information about
     *               the current protocol state.
     * @param byteBuffer the bytes response.
     */
    private final void redirectSSL(ProtocolInfo protocolInfo, ByteBuffer byteBuffer)
            throws IOException{

        SSLOutputWriter.flushChannel(protocolInfo.socketChannel,
                SC_FOUND.slice(),
                protocolInfo.outputBB,
                protocolInfo.sslEngine);
        SSLOutputWriter.flushChannel(protocolInfo.socketChannel,
                byteBuffer.slice(),
                protocolInfo.outputBB,
                protocolInfo.sslEngine);
    }
    
    
    /**
     * Redirect a un-secure request (http) to http or https.
     * @param protocolInfo the ProtocolInfo that contains the information about
     *               the current protocol state.
     */
    public final void redirect(ProtocolInfo protocolInfo)
            throws IOException{
                        
        if (!initialized){
            initializeBuffer();
        }
        
        redirect(protocolInfo,
                protocolInfo.isRequestedTransportSecure ? httpToHttps : httpToHttp);
    }
    
    
    /**
     * Redirect a secure request (http) to http or https.
     * @param protocolInfo the ProtocolInfo that contains the information about
     *               the current protocol state.
     * @param byteBuffer the bytes response.
     */
    private final void redirect(ProtocolInfo protocolInfo,ByteBuffer byteBuffer)
            throws IOException{

        OutputWriter.flushChannel(protocolInfo.socketChannel,SC_FOUND.slice());
        OutputWriter.flushChannel(protocolInfo.socketChannel,byteBuffer.slice());
    }
    
    
    /**
     * Log an exception.
     */
    private void log(Throwable ex){
        Logger logger = SSLSelectorThread.logger();
        if ( logger.isLoggable(Level.WARNING) ){
            logger.log(Level.WARNING,"Redirector",ex);
        }
    }
    
    
    public int getRedirectedPort() {
        return redirectedPort;
    }
    
    
    public void setRedirectedPort(int redirectedPort) {
        this.redirectedPort = redirectedPort;
    }

    
    public int getExternalPort() {
        return externalPort;
    }

    
    public void setExternalPort(int externalPort) {
        this.externalPort = externalPort;
    }
}
